/* interrupt.c - This file is a part of NutsOS
 *
 * NutsOS
 * Copyright (C) 2013  Free Software Foundation, Inc.
 *
 * NutsOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NutsOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with NutsOS; If not, see <http://www.gnu.org/licenses/>.
 *
 * Valentin Verdier <valentin.verdier03@gmail.com>
 */

#include "interrupt.h"

#include "types.h"
#include "asm_x86.h"
#include "print.h"
#include "process.h"

/* ============ Configuration du PIC 8259A ============ */

void PIC_Init()
{
	_out(PIC_MASTER_CMD_REG, 0x11);
	_out(PIC_SLAVE_CMD_REG, 0x11);

	_out(PIC_MASTER_DATA_REG, PIC_MASTER_IDT_OFFSET);
	_out(PIC_SLAVE_DATA_REG, PIC_SLAVE_IDT_OFFSET);

	_out(PIC_MASTER_DATA_REG, 4);
	_out(PIC_SLAVE_DATA_REG, 2);

	_out(PIC_MASTER_DATA_REG, 1);
	_out(PIC_SLAVE_DATA_REG, 1);

	_out(PIC_MASTER_DATA_REG, 0);
	_out(PIC_SLAVE_DATA_REG, 0);
}

void PIC_MaskIRQ(bits8 irq)
{
	bits8 mask;
	if(irq < 8)
	{
		mask = _in(PIC_MASTER_DATA_REG) & ~(1 << irq);
		_out(PIC_MASTER_DATA_REG, mask);
	}
	else
	{
		mask = _in(PIC_SLAVE_DATA_REG) & ~(1 << (irq - 8));
		_out(PIC_SLAVE_DATA_REG, mask);
	}
}

void PIC_Send_EOI()
{
	_out(PIC_MASTER_CMD_REG, PIC_EOI);
	_out(PIC_SLAVE_CMD_REG, PIC_EOI);
}

/* ===================================================== */

/* ========= Manipulation de l'IDT ========= */
void InitIDT()
{
	unsigned int i;
	T_IDT IdtR;
	PIC_Init();

	/* Initialisation de l'IDT */
	for(i = 0; i < IDT_SIZE; i++)
	{
		CreateInterruptDescriptor((bits32) &_DefaultInterrupt, 0x08, 0x8E00, i);
	}
	/* Configuration de l'IDT */
	CreateInterruptDescriptor((bits32) &_SEG_Fault, 0x08, 0x8E00, SEG_FAULT_INT);
	CreateInterruptDescriptor((bits32) &_SEG_Fault, 0x08, 0x8E00, STACK_SEG_FAULT_INT);

	CreateInterruptDescriptor((bits32) &_GP_Fault, 0x08, 0x8E00, GP_FAULT_INT);
	CreateInterruptDescriptor((bits32) &_Page_Fault, 0x08, 0xEF00, PAGE_FAULT_INT);

	CreateInterruptDescriptor((bits32) &_ClockInterrupt, 0x08, 0x8E00, TIMER_INT);
	CreateInterruptDescriptor((bits32) &_KeyboardInterrupt, 0x08, 0x8E00, KB_INT);

	CreateInterruptDescriptor((bits32) &_SystemCall, 0x08, 0xEF00, SYS_CALL_INT);

	/* Pointeur sur l'IDT pour lidt */
	IdtR.limite = (bits16) (INT_DESC_SIZE * IDT_SIZE);
	IdtR.base = IDT_BASE;

	_LoadIdtR(&IdtR);
}

void CreateInterruptDescriptor(bits32 offset, bits16 seg, bits16 type, unsigned int desc)
{
	T_IdtDesc *idt = (T_IdtDesc*) IDT_BASE;

	idt[desc].offset0_15 = (offset & 0xFFFF);
	idt[desc].segment = seg;
	idt[desc].type = type;
	idt[desc].offset16_31 = (offset & 0xFFFF0000) >> 16;
}
/* ===================================================== */


/* ========== Routines d'interruption en C ============ */
void IsrDefault()
{
	//print("int\n", 0x47);
	return;
}

void IsrGPFault()
{
	print("ERROR : X86 GENERAL PROTECTION FAULT\n", 0x47);
	_ClearInterrupt();
	while(1)
	{
		asm("hlt");
	}
}

void IsrSegFault()
{
	print("ERROR : X86 SEGMENTATION FAULT\n", 0x47);
	_ClearInterrupt();
	while(1)
	{
		asm("hlt");
	}
}

void IsrPageFault()
{
	bits32 cr2 = _ReadCr2();
	/*if(CurrentProcess)
	{
		ProcessAddPage(cr2 & 0xFFFFF000);
	}
	else*/
	{
		print("ERROR : PAGE FAULT (ADDR = %x)\n", 0x47, cr2);
		_ClearInterrupt();
		while(1)
		{
			asm("hlt");
		}
	}
}

void IsrClock()
{
	KBD_LedsControl(7);
	ScheduleAtClock();
}
/* ========================================================= */
