/* memory.c - This file is a part of NutsOS
 *
 * NutsOS
 * Copyright (C) 2013  Free Software Foundation, Inc.
 *
 * NutsOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NutsOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with NutsOS; If not, see <http://www.gnu.org/licenses/>.
 *
 * Valentin Verdier <valentin.verdier03@gmail.com>
 */

#define MEMORY_PRIVATE
#include "memory.h"

#include "types.h"
#include "print.h"
#include "args.h"
#include "asm_x86.h"
#include "ipc.h"

/* Initialise les principaux segments utiliss par le kernel */
void InitGDT()
{
	T_GDT GdtR;

	/* Adresse du TSS en mmoire */
	tss = (T_TSS*) TSS_BASE;

	/* Configuration du TSS */
	tss->flagDebug = 0x00;
	tss->io_map = 0x00;
	tss->esp0 = MAIN_KERNEL_STACK;
	tss->ss0 = 0x18;

	/* Configuration de la GDT */
	CreateSegDescriptor(0x0, 0x0, 0x0, 0x0, 0); /* Segment nul */
	CreateSegDescriptor(0x0, 0xFFFFF, 0x9B, 0x0D, 1); /* Code noyau */
	CreateSegDescriptor(0x0, 0xFFFFF, 0x93, 0x0D, 2); /* Donnes noyau */
	CreateSegDescriptor(0x0, 0x0, 0x97, 0x0D, 3); /* Pile noyau */

	CreateSegDescriptor(0x0, 0xFFFFF, 0xFF, 0x0D, 4); /* Code user */
	CreateSegDescriptor(0x0, 0xFFFFF, 0xF3, 0x0D, 5); /* Donnes user */
	CreateSegDescriptor(0x0, 0x0, 0xF7, 0x0D, 6); /* Pile user */

	CreateSegDescriptor((bits32) tss, 0x67, 0xE9, 0x00, 7); /* TSS */

	/* Pointeur sur GDT pour lgdt */
	GdtR.limite = (bits16) (SEG_DESC_NUMBER * SEG_DESC_SIZE);
	GdtR.base = GDT_BASE;

	_LoadGdtR(&GdtR);
	_LoadTaskRegister();
}

void CreateSegDescriptor(bits32 base, bits32 limite, bits8 acces, bits8 other, unsigned int desc)
{
	T_GdtDesc *gdt = (T_GdtDesc*) GDT_BASE;

	gdt[desc].limite0_15 = (limite & 0xFFFF);
	gdt[desc].base0_15 = (base & 0xFFFF);
	gdt[desc].base16_23 = (base & 0xFF0000) >> 16;
	gdt[desc].acces = acces;
	gdt[desc].limite16_19 = (limite & 0xF0000) >> 16;
	gdt[desc].other = (other & 0xF);
	gdt[desc].base24_31 = (base & 0xFF000000) >> 24;
}

/* Cr l'espace virtuel linaire du noyau qui couvre 4 Go */
void InitKernelSpace()
{
	bits32 i, j, page, size;

	/* === On cr l'espace d'adressage du kernel (espace physique fixe : 7184 Ko, espace virtuel dynamique : 254960 Ko, total : 256 Mo) === */

	PageDirectory = (T_PageDirectory) KERNEL_PD_ADDR;
	PageTable = (T_PageTable) KERNEL_PT_ADDR;

	/* On initialise le PageDirectory du kernel */
	for(i = 0; i < MAX_PT_COUNT; i++)
	{
		if(i < KERNEL_SPACE_PT_COUNT)
		{
			PageDirectory[i] = KERNEL_PT_ADDR + i * PT_SIZE;
			PageDirectory[i] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW;
		}
		else
		{
			PageDirectory[i] = 0;
		}
	}

	/* On initialise les tables de pages du kernel */
	for(i = 0; i < KERNEL_SPACE_PAGE_COUNT; i++)
	{
		if(i < KERNEL_PHY_SPACE_PAGE_COUNT)
		{
			PageTable[i] = i * PAGE_SIZE;
			PageTable[i] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW;
		}
		else
		{
			PageTable[i] = 0;
		}
	}

	/* ============================================== */



	/* ==== On initialise la table d'tat des pages physiques en utilisant la Memory Map ==== */

	PageMap = (bits8*) PHY_PAGE_MAP_ADDR;

	for(i = 0; i < MAX_PAGE_COUNT; i++) /* Initialisation */
	{
		PageMap[i] = PAGE_NOT_PRESENT;
	}
	/* On parcout la Memory Map */
	for(i = 0; i < arg->memListEntryNumber; i++)
	{
		/* On calcul le numro de la page de base puis la taille en nombre de page */
		page = arg->memMap[i].baseAddr_1 / PAGE_SIZE;
		size = arg->memMap[i].size_1 / PAGE_SIZE;

		if(arg->memMap[i].size_1 % PAGE_SIZE > 0)
		{
			size++;
		}

		/* Si la zone mmoire est utilisable */
		if(arg->memMap[i].type == 1)
		{
			/* On parcourt la section de la table d'tat correspondant  la zone mmoire */
			for(j = page; j < page + size; j++)
			{
				PageMap[j] = FREE;
			}
		}
	}

	/* =============================================================== */


	/* ===== Structures d'allocation ===== */

	KERNEL_HEAP_BASE = KERNEL_PHY_SPACE_PAGE_COUNT << PAGE_OFFSET_BITS;
	if(!KernelAddPage(KERNEL_HEAP_BASE))
	{
		fatalError("MEGA ERROR OF DEATH !!");
	}
	KERNEL_HEAP_LIMIT = KERNEL_HEAP_BASE + PAGE_SIZE;

	struct AllocAreaModel *firstArea = (struct AllocAreaModel*) KERNEL_HEAP_BASE;
	firstArea->state = 0;
	firstArea->next = 0;

	/* ================================================================= */


	PhyPageMapMutex = 0;
	KernPageTableMutex = 0;
	AllocMutex = 0;
}

unsigned int KernelAddPage(T_Page addr)
{
	unsigned int vpage = addr >> PAGE_OFFSET_BITS;
	unsigned int ret;

	if(vpage < KERNEL_SPACE_PAGE_COUNT && vpage >= KERNEL_PHY_SPACE_PAGE_COUNT)
	{
		IPC_MutexLock(&KernPageTableMutex);
		if(PageTable[vpage] == 0)
		{
			PageTable[vpage] = GetPhyPage();
			if(PageTable[vpage])
			{
				PageTable[vpage] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW;
				ret = 1;
			}
			else
			{
				ret = 0;
			}
		}
		else
		{
			print("[KernelAddPage] vaddr = 0x%x is already used\n", 0x47, addr);
			ret = 0;
		}
		IPC_MutexUnlock(&KernPageTableMutex);
	}
	else
	{
		print("[KernelAddPage] vaddr = 0x%x isn't in virtual kernel space\n", 0x47, addr);
		ret = 0;
	}
	return ret;
}

void KernelRemovePage(T_Page vaddr)
{
	unsigned int vpage = vaddr >> PAGE_OFFSET_BITS;

	if(vpage < KERNEL_SPACE_PAGE_COUNT && vpage >= KERNEL_PHY_SPACE_PAGE_COUNT)
	{
		IPC_MutexLock(&KernPageTableMutex);
		if(PageTable[vpage] != 0)
		{
			ReleasePhyPage(PageTable[vpage] & 0xFFFFF000);
			PageTable[vpage] = 0;
			_EmptyTLB();
		}
		else
		{
			print("[KernelRemovePage] vaddr = 0x%x is already released\n", 0x47);
		}
		IPC_MutexUnlock(&KernPageTableMutex);
	}
	else
	{
		print("[KernelRemovePage] vaddr = 0x%x isn't in virtual kernel space\n", 0x47);
	}
}

T_Page KernelGetPage()
{
	unsigned int i;
	T_Page ret;

	for(i = KERNEL_SPACE_PAGE_COUNT - 1; i >= KERNEL_HEAP_LIMIT >> PAGE_OFFSET_BITS; i--)
	{
		IPC_MutexLock(&KernPageTableMutex);
		if(PageTable[i] == 0)
		{
			PageTable[i] = GetPhyPage();
			if(PageTable[i])
			{
				PageTable[i] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW;
				ret = i << PAGE_OFFSET_BITS;
			}
			else
			{
				ret = 0;
			}
			IPC_MutexUnlock(&KernPageTableMutex);
			break;
		}
		else
		{
			ret = 0;
		}
		IPC_MutexUnlock(&KernPageTableMutex);
	}
	return ret;
}

void* KernelGetPhyAddr(void *vaddr)
{
	unsigned int vpage = (unsigned int) vaddr >> PAGE_OFFSET_BITS;
	void *ret;

	if(vpage < KERNEL_SPACE_PAGE_COUNT && vpage >= KERNEL_PHY_SPACE_PAGE_COUNT)
	{
		IPC_MutexLock(&KernPageTableMutex);
		if(PageTable[vpage])
		{
			ret = (void*) (PageTable[vpage] & 0xFFFFF000 | (unsigned int) vaddr & 0x00000FFF);
		}
		else
		{
			ret = 0;
		}
		IPC_MutexUnlock(&KernPageTableMutex);
	}
	else
	{
		ret = 0;
	}
	return ret;
}

T_Page GetPhyPage()
{
	unsigned int i;
	for(i = KERNEL_PHY_SPACE_PAGE_COUNT; i < MAX_PAGE_COUNT; i++)
	{
		IPC_MutexLock(&PhyPageMapMutex);
		if(PageMap[i] == FREE)
		{
			PageMap[i] = USED;
			IPC_MutexUnlock(&PhyPageMapMutex);
			return i << PAGE_OFFSET_BITS;
		}
		IPC_MutexUnlock(&PhyPageMapMutex);
	}
	print("[GetPhyPage] no enough physical pages\n", 0x47);
	return 0;
}

void ReleasePhyPage(T_Page addr)
{
	unsigned int page = addr >> PAGE_OFFSET_BITS;

	if(page >= KERNEL_PHY_SPACE_PAGE_COUNT)
	{
		IPC_MutexLock(&PhyPageMapMutex);
		if(PageMap[page] != PAGE_NOT_PRESENT)
		{
			PageMap[page] = FREE;
		}
		else
		{
			print("[ReleasePhyPage] unable to release a not present page\n", 0x47);
		}
		IPC_MutexUnlock(&PhyPageMapMutex);
	}
	else
	{
		print("[ReleasePhyPage] unable to release a kernel physical page\n", 0x47);
	}
}

void* malloc(unsigned int size)
{
	IPC_MutexLock(&AllocMutex);

	struct AllocAreaModel *area = (struct AllocAreaModel*) KERNEL_HEAP_BASE;
	void *ptr = 0;

	start:

	while(area != 0)
	{
		if(((unsigned int) area + sizeof(struct AllocAreaModel) < KERNEL_HEAP_LIMIT))
		{
			if(area->next >= (struct AllocAreaModel*) KERNEL_HEAP_LIMIT)
			{
				fatalError("MEMORY MANAGMENT DATA STRUCTURES ARE DAMAGED");
			}

			if(area->state == FREE)
			{
				unsigned int areaBase = (unsigned int) area + sizeof(struct AllocAreaModel);
				unsigned int areaSize;

				if(area->next == 0)
				{
					areaSize = KERNEL_HEAP_LIMIT - areaBase;
				}
				else
				{
					areaSize = (unsigned int) area->next - areaBase;
				}

				if(areaSize == size)
				{
					area->state = USED;
					ptr = (void*) areaBase;
					break;
				}
				else if(areaSize > size)
				{
					struct AllocAreaModel *newArea = (struct AllocAreaModel*) (areaBase + size);

					if(newArea + sizeof(struct AllocAreaModel) < (struct AllocAreaModel*) KERNEL_HEAP_LIMIT)
					{
						newArea->state = FREE;
						newArea->next = area->next;
						area->next = newArea;
					}
					else
					{
						area->next = 0;
					}
					area->state = USED;
					ptr = (void*) areaBase;
					break;
				}
			}
		}
		else
		{
			fatalError("MEMORY MANAGMENT DATA STRUCTURES ARE DAMAGED");
		}
		area = area->next;
	}

	if(!ptr)
	{
		area = EnlargeHeap();
		if(area)
		{
			goto start;
		}
	}

	/*print("MALLOC -> ", 0x04);
	area = KERNEL_HEAP_BASE;
	while(area)
	{
		print("base : 0x%x, next : 0x%x, state : %d\n", 0x04, (unsigned int) area + sizeof(struct AllocAreaModel), area->next, area->state);
		area = area->next;
	}*/

	IPC_MutexUnlock(&AllocMutex);
	return ptr;
}

void free(void *addr)
{
	IPC_MutexLock(&AllocMutex);

	struct AllocAreaModel *area = (struct AllocAreaModel*) KERNEL_HEAP_BASE;

	while(area != 0)
	{
		if(area->next >= (struct AllocAreaModel*) KERNEL_HEAP_LIMIT)
		{
			fatalError("MEMORY MANAGMENT DATA STRUCTURES ARE DAMAGED");
		}

		if(area->state == USED)
		{
			unsigned int areaBase = (unsigned int) area + sizeof(struct AllocAreaModel);
			if(areaBase == (unsigned int) addr)
			{
				area->state = FREE;

				while(area->next != 0)
				{
					if(area->next >= (struct AllocAreaModel*) KERNEL_HEAP_LIMIT)
					{
						fatalError("MEMORY MANAGMENT DATA STRUCTURES ARE DAMAGED");
					}

					if(area->next->state == FREE)
					{
						area->next = area->next->next;
					}
					else
					{
						break;
					}
				}

				IPC_MutexUnlock(&AllocMutex);
				return;
			}
		}
		area = area->next;
	}

	/*print("FREE -> ", 0x04);
	area = KERNEL_HEAP_BASE;
	while(area)
	{
		print("base : 0x%x, next : 0x%x, state : %d\n", 0x04, (unsigned int) area + sizeof(struct AllocAreaModel), area->next, area->state);
		area = area->next;
	}*/
	fatalError("IMPOSSIBLE TO FREE MEMORY");
	IPC_MutexUnlock(&AllocMutex);
}

struct AllocAreaModel* EnlargeHeap()
{
	if(KernelAddPage(KERNEL_HEAP_LIMIT))
	{
		struct AllocAreaModel *area = (struct AllocAreaModel*) KERNEL_HEAP_BASE;
		while(area->next != 0)
		{
			area = area->next;
		}

		struct AllocAreaModel *lastArea = area;

		if(area->state == USED)
		{
			lastArea = (struct AllocAreaModel*) KERNEL_HEAP_LIMIT;
			lastArea->next = 0;
			lastArea->state = FREE;

			area->next = lastArea;
		}
		KERNEL_HEAP_LIMIT += PAGE_SIZE;
		return lastArea;
	}
	else
	{
		print("[EnlargeHeap] unable to enlarge kernel heap, no enough kernel pages\n", 0x47);
		return 0;
	}
}

T_PageDirectory CreateTaskPageDirectory()
{
	/* #-----# 0xFFFFFFFF
	 * |-----| pageDir user
	 * |     |
	 * |     |
	 * |     |
	 * |     |
	 * |     |
	 * |-----|
	 * |     |
	 * #-----# espace noyau */


	unsigned int i;
	T_PageDirectory PageDir = (T_PageDirectory) KernelGetPage();

	if(PageDir)
	{
		for(i = 0; i < MAX_PT_COUNT; i++) /* Initialisation du rpertoire de pages*/
		{
			PageDir[i] = 0;
		}

		for(i = 0; i < KERNEL_SPACE_PT_COUNT; i++) /* Construction de la partie noyau */
		{
			PageDir[i] = PageDirectory[i];
		}

		/* La table de pages 1023 couvre le PageDir et les table de pages, de 0xFFC00000  0xFFFFFFFF */

		PageDir[1023] = (T_Page) KernelGetPhyAddr(PageDir);
		PageDir[1023] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW;
		return PageDir;
	}
	else
	{
		return 0;
	}
}

void DestroyTaskPageDirectory(T_PageDirectory PageDir)
{
	unsigned int i;
	for(i = KERNEL_SPACE_PAGE_COUNT; i < MAX_PT_COUNT; i++)
	{
		if(PageDir[i])
		{
			ReleasePhyPage(PageDir[i] & 0xFFFFF000);
		}
	}

	KernelRemovePage((T_Page) PageDir);
}

/* Alloue une nouvelle page physique et la fait correspondre  l'adresse virtuelle vaddr */
unsigned int AddPage(T_Page vaddr, bits32 perm)
{
	unsigned int i;

	T_PageDirectory PageDir = (T_PageDirectory) 0xFFFFF000;
	T_PageTable PageTableAddr;
	T_PageTable PageTableVaddr;
	bits32 ptNum = (vaddr & 0xFFC00000) >> 22;
	bits32 pgNum = (vaddr & 0x003FF000) >> 12;

	/* On regarde si la table de pages est cre */
	if(PageDir[ptNum] == 0) /* Si elle n'xiste pas */
	{
		/* On cr la table de pages */
		PageTableAddr = (T_PageTable) GetPhyPage(); /* Adresse physique, donc inaccessible ici */
		if(PageTableAddr)
		{
			/* On ajoute la table de pages cre dans le rpertoire */
			PageDir[ptNum] = (T_Page) PageTableAddr;
			PageDir[ptNum] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW | PAGE_FLAG_USER;

			PageTableVaddr = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE); /* On accde  la table de pages depuis l'espace de mappage virtuel */

			for(i = 0; i < 1024; i++)
			{
				PageTableVaddr[i] = 0;
			}

			PageTableVaddr[pgNum] = GetPhyPage();
			if(PageTableVaddr[pgNum])
			{
				PageTableVaddr[pgNum] |= perm;
				return 1;
			}
			else
			{
				PageDir[ptNum] = 0;
				ReleasePhyPage((T_Page) PageTableAddr);
				_EmptyTLB();
				return 0;
			}
		}
		else
		{
			return 0;
		}

	}
	else /* Si elle xiste */
	{
		PageTableVaddr = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE); /* On accde  la table de pages depuis l'espace de mappage virtuel */

		/* On test pour savoir si la page voulue est dj prsente */
		if(PageTableVaddr[pgNum] == 0)
		{
			/* Si elle ne l'est pas on ajoute  la table de pages la page voulue */
			PageTableVaddr[pgNum] = GetPhyPage();
			if(PageTableVaddr[pgNum])
			{
				PageTableVaddr[pgNum] |= perm;
				return 1;
			}
			else
			{
				return 0;
			}
		}
		else
		{
			return 0;
		}
	}
}

void RemovePage(T_Page vaddr)
{
	unsigned int i;

	T_PageDirectory PageDir = (T_PageDirectory) 0xFFFFF000;
	T_PageTable PageTable;
	int empty;
	bits32 ptNum = (vaddr & 0xFFC00000) >> 22;
	bits32 pgNum = (vaddr & 0x003FF000) >> 12;

	if(PageDir[ptNum] != 0)
	{
		PageTable = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE);
		if(PageTable[pgNum] != 0)
		{
			ReleasePhyPage(PageTable[pgNum] & 0xFFFFF000);
			PageTable[pgNum] = 0;

			for(i = 0; i < PAGE_PER_PT; i++)
			{
				if(PageTable[i] == 0)
				{
					empty = 1;
				}
				else
				{
					empty = 0;
					break;
				}
			}

			if(empty)
			{
				ReleasePhyPage(PageDir[ptNum] & 0xFFFFF000);
				PageDir[ptNum] = 0;
			}
			_EmptyTLB();
		}
	}
}

unsigned int LinkPhyPage(T_Page vaddr, T_Page addr, bits32 perm)
{
	unsigned int i;

	T_PageDirectory PageDir = (T_PageDirectory) 0xFFFFF000;
	T_PageTable PageTableAddr;
	T_PageTable PageTableVaddr;
	bits32 ptNum = (vaddr & 0xFFC00000) >> 22;
	bits32 pgNum = (vaddr & 0x003FF000) >> 12;

	/* On regarde si la table de pages est cre */
	if(PageDir[ptNum] == 0) /* Si elle n'xiste pas */
	{
		/* On cr la table de pages */
		PageTableAddr = (T_PageTable) GetPhyPage(); /* Adresse physique, donc inaccessible ici */
		if(PageTableAddr)
		{
			/* On ajoute la table de pages cre dans le rpertoire */
			PageDir[ptNum] = (T_Page) PageTableAddr;
			PageDir[ptNum] |= PAGE_FLAG_PRESENT | PAGE_FLAG_RW | PAGE_FLAG_USER;

			PageTableVaddr = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE); /* On accde  la table de pages depuis l'espace de mappage virtuel */

			for(i = 0; i < 1024; i++)
			{
				PageTableVaddr[i] = 0;
			}

			PageTableVaddr[pgNum] = addr & 0xFFFFF000;
			PageTableVaddr[pgNum] |= perm;
			return 1;
		}
		else
		{
			return 0;
		}

	}
	else /* Si elle xiste */
	{
		PageTableVaddr = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE); /* On accde  la table de pages depuis l'espace de mappage virtuel */

		/* On test pour savoir si la page voulue est dj prsente */
		if(PageTableVaddr[pgNum] == 0)
		{
			/* Si elle ne l'est pas on ajoute  la table de pages la page voulue */
			PageTableVaddr[pgNum] = addr & 0xFFFFF000;
			PageTableVaddr[pgNum] |= perm;
			return 1;
		}
		else
		{
			return 0;
		}
	}
}

void UnlinkPhyPage(T_Page vaddr)
{
	unsigned int i;

	T_PageDirectory PageDir = (T_PageDirectory) 0xFFFFF000;
	T_PageTable PageTable;
	int empty;
	bits32 ptNum = (vaddr & 0xFFC00000) >> 22;
	bits32 pgNum = (vaddr & 0x003FF000) >> 12;

	if(PageDir[ptNum] != 0)
	{
		PageTable = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE);
		if(PageTable[pgNum] != 0)
		{
			PageTable[pgNum] = 0;
			for(i = 0; i < PAGE_PER_PT; i++)
			{
				if(PageTable[i] == 0)
				{
					empty = 1;
				}
				else
				{
					empty = 0;
					break;
				}
			}

			if(empty)
			{
				ReleasePhyPage(PageDir[ptNum] & 0xFFFFF000);
				PageDir[ptNum] = 0;
			}
			_EmptyTLB();
		}
	}
}

void* GetPhyAddr(void *vaddr)
{
	T_PageDirectory PageDir = (T_PageDirectory) 0xFFFFF000;
	T_PageTable PageTable;
	bits32 ptNum = ((bits32) vaddr & 0xFFC00000) >> 22;
	bits32 pgNum = ((bits32) vaddr & 0x003FF000) >> 12;
	bits32 offset = (bits32) vaddr & 0x00000FFF; /* Offset par rapport au dbut de la page */

	if(PageDir[ptNum] != 0)
	{
		PageTable = (T_PageTable) (0xFFC00000 + ptNum * PT_SIZE);
		if(PageTable[pgNum] != 0)
		{
			return (void*) (PageTable[pgNum] & 0xFFFFF000 | offset);
		}
	}

	return 0;
}
