/* process.c - This file is a part of NutsOS
 *
 * NutsOS
 * Copyright (C) 2013  Free Software Foundation, Inc.
 *
 * NutsOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NutsOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with NutsOS; If not, see <http://www.gnu.org/licenses/>.
 *
 * Valentin Verdier <valentin.verdier03@gmail.com>
 */

#define PROCESS_PRIVATE
#include "process.h"

#include "types.h"
#include "display.h"
#include "asm_x86.h"
#include "memory.h"
#include "utils.h"
#include "args.h"
#include "elf_loader.h"

void Process_0_Task();


void InitProcessManager()
{
	Process_0.pid = 0;
	Process_0.reg_EAX = 0x00;
	Process_0.reg_EBX = 0x00;
	Process_0.reg_ECX = 0x00;
	Process_0.reg_EDX = 0x00;
	Process_0.reg_ESI = 0x00;
	Process_0.reg_EDI = 0x00;
	Process_0.reg_EBP = 0x00;
	Process_0.reg_ESP = MAIN_KERNEL_STACK;
	Process_0.reg_CS = 0x08;
	Process_0.reg_DS = 0x10;
	Process_0.reg_ES = 0x10;
	Process_0.reg_FS = 0x10;
	Process_0.reg_GS = 0x10;
	Process_0.reg_SS = 0x18;
	Process_0.reg_EIP = (bits32) &Process_0_Task;
	Process_0.reg_EFLAGS = 0x200;
	Process_0.pageDir = (T_PageDirectory) KERNEL_PD_ADDR;
	Process_0.reg_CR3 = KERNEL_PD_ADDR;
	Process_0.kernelStackPtr = MAIN_KERNEL_STACK;

	/* Pas de mode user, donc pas de régions */
	Process_0.processRegionsList.firstEntry = 0;
	Process_0.processRegionsList.lastEntry = 0;
	Process_0.processRegionsList.entryCount = 0;

	Process_0.parent = 0; /* Pas de parent */
	Process_0.processChildrenList.firstEntry = 0;
	Process_0.processChildrenList.lastEntry = 0;
	Process_0.processChildrenList.entryCount = 0;

	Process_0.signal = 0;

	Process_0.status = PROCESS_STATUS_RUNNING;
	Process_0.ipcWaitMutex = 0;
	Process_0.ipcInSection = 0;

	Processes.firstEntry = 0;
	Processes.lastEntry = 0;
	Processes.entryCount = 0;
	LinkedListPushBack(&Process_0, &Processes);

	PidCounter = 1;
}

unsigned int CreateProcess(void *file)
{
	unsigned int i;

	/* On initialise une entrée dans la table des processus */
	T_Process* process = malloc(sizeof(T_Process));
	struct ProcessChild* child = malloc(sizeof(struct ProcessChild));
	if(!process || !child)
	{
		if(child)
		{
			free(child);
		}
		if(process)
		{
			free(process);
		}
		print("[CreateProcess] unable to create process, not enough kernel memory\n", 0x47);
		return 0;
	}

	process->pid = PidCounter;
	process->status = PROCESS_STATUS_RUNNING;
	process->ipcWaitMutex = 0;
	process->ipcInSection = 0;

	process->processRegionsList.firstEntry = 0;
	process->processRegionsList.lastEntry = 0;
	process->processRegionsList.entryCount = 0;

	child->process = process;
	process->parent = CurrentProcess;
	process->processChildrenList.firstEntry = 0;
	process->processChildrenList.lastEntry = 0;
	process->processChildrenList.entryCount = 0;

	process->signal = 0;

	process->reg_EAX = 0x00;
	process->reg_EBX = 0x00;
	process->reg_ECX = 0x00;
	process->reg_EDX = 0x00;
	process->reg_ESI = 0x00;
	process->reg_EDI = 0x00;
	process->reg_EBP = 0x00;
	process->reg_ESP = (bits32) USER_STACK_ADDR + PAGE_SIZE;
	process->reg_CS = 0x23;
	process->reg_DS = 0x2B;
	process->reg_ES = 0x2B;
	process->reg_FS = 0x2B;
	process->reg_GS = 0x2B;
	process->reg_SS = 0x33;
	process->reg_EFLAGS = 0x200;
	/* ======================================================= */

	/* On créé le répertoire de pages */
	process->pageDir = CreateTaskPageDirectory();
	if(!process->pageDir)
	{
		free(process);
		free(child);
		print("[CreateProcess] unable to create process, not enough kernel memory\n", 0x47);
		return 0;
	}

	process->reg_CR3 = (bits32) KernelGetPhyAddr(process->pageDir);
	/* ================================================= */

	/* La pile noyau */
	process->kernelStackPtr = KernelGetPage();
	if(!process->kernelStackPtr)
	{
		DestroyTaskPageDirectory(process->pageDir);
		free(process);
		free(child);
		print("[CreateProcess] unable to create process, not enough kernel memory\n", 0x47);
		return 0;
	}
	process->kernelStackPtr += PAGE_SIZE;
	/* ================================================== */

	/* On change d'espace d'adressage */
	bits32 cr3 = _ReadCr3();
	CurrentProcess->reg_CR3 = process->reg_CR3;
	_LoadCr3(process->reg_CR3);


	/* On créé la pile utilisateur */
	if(!AddPage((T_Page) USER_STACK_ADDR, PAGE_FLAG_PRESENT | PAGE_FLAG_RW | PAGE_FLAG_USER))
	{
		CurrentProcess->reg_CR3 = cr3;
		_LoadCr3(cr3);
		KernelRemovePage(process->kernelStackPtr - PAGE_SIZE);
		DestroyTaskPageDirectory(process->pageDir);
		free(process);
		free(child);
		print("[CreateProcess] unable to create process, not enough memory\n", 0x47);
		return 0;
	}
	/* ====================================================== */

	process->reg_EIP = ELF32_Load(file, &process->processRegionsList);
	if(!process->reg_EIP)
	{
		RemovePage((T_Page) USER_STACK_ADDR);
		CurrentProcess->reg_CR3 = cr3;
		_LoadCr3(cr3);
		KernelRemovePage(process->kernelStackPtr - PAGE_SIZE);
		DestroyTaskPageDirectory(process->pageDir);
		free(process);
		free(child);
		print("[CreateProcess] unable to create process, elf error\n", 0x47);
		return 0;
	}

	/* On restaure l'espace d'adressage */
	CurrentProcess->reg_CR3 = cr3;
	_LoadCr3(cr3);

	/* On met à jour les tables */
	PidCounter++;
	LinkedListPushBack(child, &CurrentProcess->processChildrenList);
	LinkedListPushBack(process, &Processes);

	print("[CreateProcess] new process created with PID %d\n", DEFAULT_COLOR, process->pid);
	return 1;
}


void DestroyProcess()
{

	/*unsigned int i;
	for(i = 0; i < ProcessNumber; i++)
	{
		struct ProcessModel *process = GetEntry(i, &Processes);
		if(process == CurrentProcess)
		{
			unsigned int j;
			print("Stopping process with PID %d\n", DEFAULT_COLOR, CurrentProcess->pid);

			for(j = 0; j < CurrentProcess->codePageCount; j++)
			{
				RemovePage((bits32) USER_SPACE_BASE_ADDR + (j * PAGE_SIZE));
			}


			for(j = 0; j < CurrentProcess->allocatedPageCount; j++)
			{
				struct ProcessAllocatedPageModel *page = GetEntry(0, &CurrentProcess->allocatedPageList);
				RemovePage(page->page_vaddr);
				free(page);
				RemoveEntry(0, &CurrentProcess->allocatedPageList);
			}

			RemovePage((bits32) USER_STACK_ADDR);
			CurrentProcess->status = PROCESS_STATUS_ZOMBIE;
			while(1)
			{
				asm("hlt");
			}
			KernelRemovePage(CurrentProcess->kernelStackPtr - PAGE_SIZE);
			DestroyTaskPageDirectory(CurrentProcess->pageDir);
			free(process);


			bits32 eflags = _SaveEflags();
			_ClearInterrupt();
			RemoveEntry(i, &Processes);
			ProcessNumber--;
			_RestoreEflags(eflags)
		}
	}*/
}

void Process_0_Task()
{
	print("[God] create Init process with PID 1\n", DEFAULT_COLOR);
	CreateProcess((void*) arg->initProcessAddr);

	while(1)
	{
		asm("hlt");
	}
}


unsigned int ProcessCreateRegion_RO(T_Page startPage, T_Page* pglist, unsigned int pageCount, T_LinkedList* regionsList)
{
	startPage &= 0xFFFFF000;
	T_Page endPage = startPage + (pageCount << PAGE_OFFSET_BITS);

	if(!pageCount || startPage < USER_SPACE_BASE_ADDR || startPage >= USER_SPACE_END_ADDR || endPage >= USER_SPACE_END_ADDR)
	{
		print("[ProcessCreateRegion_RO] invalid region location\n", 0x47);
		return 0;
	}

	struct ProcessRegion* entry = (struct ProcessRegion*) regionsList->firstEntry;
	while(entry)
	{
		if(endPage > entry->startPage && endPage <= entry->endPage
		|| startPage >= entry->startPage && startPage < entry->endPage
		|| startPage <= entry->startPage && endPage >= entry->endPage)
		{
			print("[ProcessCreateRegion_RO] conflict between regions\n", 0x47);
			return 0;
		}
		entry = entry->next;
	}

	unsigned int i, pc;
	for(i = 0, pc = 0; i < pageCount; i++, pc++)
	{
		if(!LinkPhyPage(startPage + (i << PAGE_OFFSET_BITS), pglist[i] & 0xFFFFF000, PAGE_FLAG_PRESENT | PAGE_FLAG_USER))
		{
			goto error;
		}
	}

	struct ProcessRegion* region = malloc(sizeof(struct ProcessRegion));
	if(region)
	{
		region->startPage = startPage;
		region->endPage = endPage;
		region->type = (unsigned char) PROCESS_REGION_RO;
		LinkedListPushBack(region, regionsList);
		return 1;
	}

	error:
		print("[ProcessCreateRegion_RO] unable to add region\n", 0x47);
		for(i = 0; i < pc; i++)
		{
			UnlinkPhyPage(startPage + (i << PAGE_OFFSET_BITS));
		}
		return 0;
}

unsigned int ProcessCreateRegion_RW(T_Page startPage, unsigned int pageCount, T_LinkedList* regionsList)
{
	startPage &= 0xFFFFF000;
	T_Page endPage = startPage + (pageCount << PAGE_OFFSET_BITS);

	if(!pageCount || startPage < USER_SPACE_BASE_ADDR || startPage >= USER_SPACE_END_ADDR || endPage >= USER_SPACE_END_ADDR)
	{
		print("[ProcessCreateRegion_RW] invalid region location\n", 0x47);
		return 0;
	}

	struct ProcessRegion* entry = (struct ProcessRegion*) regionsList->firstEntry;
	while(entry)
	{
		if(endPage > entry->startPage && endPage <= entry->endPage
		|| startPage >= entry->startPage && startPage < entry->endPage
		|| startPage <= entry->startPage && endPage >= entry->endPage)
		{
			print("[ProcessCreateRegion_RW] conflict between regions\n", 0x47);
			return 0;
		}
		entry = entry->next;
	}

	unsigned int i, pc;
	for(i = 0, pc = 0; i < pageCount; i++, pc++)
	{
		if(!AddPage(startPage + (i << PAGE_OFFSET_BITS), PAGE_FLAG_PRESENT | PAGE_FLAG_RW | PAGE_FLAG_USER))
		{
			goto error;
		}
	}

	struct ProcessRegion* region = malloc(sizeof(struct ProcessRegion));
	if(region)
	{
		region->startPage = startPage;
		region->endPage = endPage;
		region->type = (unsigned char) PROCESS_REGION_RW;
		LinkedListPushBack(region, regionsList);
		return 1;
	}

	error:
		print("[ProcessCreateRegion_RW] unable to add region\n", 0x47);
		for(i = 0; i < pc; i++)
		{
			RemovePage(startPage + (i << PAGE_OFFSET_BITS));
		}
		return 0;
}

void ProcessRemoveRegion(T_Page startPage, unsigned int pageCount, T_LinkedList* regionsList)
{
	startPage &= 0xFFFFF000;
	T_Page endPage = startPage + (pageCount << PAGE_OFFSET_BITS);

	struct ProcessRegion* entry = (struct ProcessRegion*) regionsList->firstEntry;
	while(entry)
	{
		if(startPage == entry->startPage && endPage == entry->endPage)
		{
			unsigned int i;
			if(entry->type == PROCESS_REGION_RO)
			{
				for(i = 0; i < pageCount; i++)
				{
					UnlinkPhyPage(startPage + (i << PAGE_OFFSET_BITS));
				}
				LinkedListRemove(entry, regionsList);
				free(entry);
				break;
			}
			else if(entry->type == PROCESS_REGION_RW)
			{
				for(i = 0; i < pageCount; i++)
				{
					RemovePage(startPage + (i << PAGE_OFFSET_BITS));
				}
				LinkedListRemove(entry, regionsList);
				free(entry);
				break;
			}
		}
		entry = entry->next;
	}
}
